# -*- coding: utf-8 -*-
from calfbox import cbox #calfbox is a globally installed module.

cbox.init_engine()
cbox.Config.set("io", "client_name", "Laborejo")
cbox.start_audio()

"""This file creates classes that are compatible with libsmf calls.
You can override the default libsmf instances and functions to
generate live output through calfbox instead. Only use it temporary
though.
Laborejo expects libsmf to be the standard playback module.
In the future we might even change the smf module and classes
to be a parameter.

There is a module check in items and core to see if cbox is installed.
cboxwrap is only imported after that. If you want to import from
another place you need to check if cbox exists. Laborejo should run
without any playback features."""

class Calfbox(object):
    """Drop in replacement for libsmf smf, the module directly.
    We can't name calfboxwrapper.py smf because that would lead to
    conflicts. And I also don't want to overwrite and reload modules."""
    def __init__(self, ppqn, soundfont, autoconnect = True, midiInPort = None):
        self._SMF = self.Smf(ppqn)
        self.transport = cbox.Transport
        #cbox.do_cmd("/master/set_tempo", None, [120]) # Set tempo
        self.scene = self._SMF.calfBoxDocument.get_scene()
        self.newInternalInstrument(soundfont)
        self.cbox = cbox
        self.autoconnect = autoconnect
        self.jack_midi_in_is_open = True #A switch to 'mute' midi in for Laborejo without destroying any connections. The midi in data is just discarded.
        self.states = States()
        self.executeCC = {} #populate with key[CCnumber]:function(midiVelocity). If there is any controller change.


        #If the connection below fails for some reason calfbox will raise an exception eventhough nothing critical happened. We need to catch that and ignore it.
        #if self.autoconnect: ##NO. That is the only thing that is for pure convenience. Auto-Connection will always happen here.
        try:
            outputs = cbox.JackIO.get_ports(".*", cbox.JackIO.AUDIO_TYPE, cbox.JackIO.PORT_IS_SINK | cbox.JackIO.PORT_IS_PHYSICAL)
            cbox.JackIO.port_connect("Laborejo:out_1", outputs[0]) #system:playback_1
            cbox.JackIO.port_connect("Laborejo:out_2", outputs[1]) #system:playback_2
        except:
            pass

        if self.autoconnect and midiInPort:
            try:
                cbox.JackIO.port_connect(midiInPort, "Laborejo:midi") #maybe that is already connected or the port was wrong.
            except:
                pass

    def __deepcopy__(self, memo):
        """Just a precaution. Don't copy anything. Just return itself."""
        return self

    def newInternalInstrument(self, filepath):
        self.scene.clear()
        self.instrument = self.scene.add_new_instrument_layer("internalGeneralMidi", "fluidsynth").get_instrument()
        self.instrument.cmd("/engine/load_soundfont", None, filepath)

    def Event(self, event):
        """libsmf converts to binary when creating an event. Here we just
        return what we got."""
        return event

    def SMF(self, ppqn):
        """ppqn is ignored. It is 384 laborejo wide anyway"""
        return self._SMF

    def play(self, startTick):
        self.stop()
        self.transport.seek_ppqn(startTick)
        self.transport.play()

    def stop(self):
        self.transport.stop()

    def panic(self):
        """Also does stop"""
        self.transport.panic()

    def connectMidiOutput(self, uuid, externalPort):
        """A wrapper to test if autoconnection is on or off.
        Session managers don't like auto-connection.
        externalPort is a string "x:y"""
        if self.autoconnect:
            cbox.JackIO.autoconnect_midi_output(uuid, externalPort)

    class Smf(object):
        """Drop-In for smfStructure. used in Laborejo as smf.SMF(384)
         The whole midi document"""
        def __init__(self, ppqn):
            self.ppqn = ppqn
            #song and scene are singletons within a document, and get_song() returns the song singleton
            self.calfBoxDocument = cbox.Document
            #self.calfBoxDocument.dump() #gives you the list of all objects in the document, so it might be used to debug some object lifecycle issues
            self.song = self.calfBoxDocument.get_song()
            self.song.clear() # Delete all the tracks and patterns #TODO. Really necessary for a new song?
            #self.song.set_loop(pattern_len, pattern_len) # Stop the song at the end
            self.tracks = []
            self.transport = cbox.Transport
            self.cbox = cbox
            self.lastMaxTicks = None
            #master = smfStructure.transport.status() #'pos', 'pos_ppqn', 'tempo', 'timesig', 'sample_rate'


        def __deepcopy__(self, memo):
            """Just a precaution. Don't copy anything. Just return itself."""
            return self

        def save(self, filepath):
            """used by the api functions
            Is not implemented and probably will never be used
            since we have libsmf to create midi files.
            Calling this functions means you tried to export midi
            and somehow ended up using calfbox. Error. """
            raise RuntimeError("Attempt to save a midi file thru Calfbox. This is not possible and should not have happened. Only save to files with the libsmf module""")

        def add_track(self, backendTrack):
            """Add a new track to the midi structure"""
            cboxTr = CalfboxTrack(backendTrack)
            self.tracks.append(cboxTr)
            return cboxTr

        def update(self, jackMode):
            """take the currently cached CalfboxTracks and convert them
            into realtime data
            """
            lengths = []
            self.song.clear()
            for cboxTr in self.tracks:
                binaryTrack = self.song.add_track()
                if jackMode:
                    binaryTrack.set_external_output(cboxTr.backendTrack.calfboxUuid) #this replaces the internal routing as well. So we don't need to deactivate sf2 output manually. Has NOTHING to do with external synth/sampler connections.
                # Create a new pattern object using events from the blob
                pattern = self.song.pattern_from_blob(cboxTr.pblob, cboxTr.length)
                lengths.append(cboxTr.length)

                # Add an instance (clip) of the pattern to the track at position 0
                # The clip will contain the whole pattern (it is also possible to insert
                # a single slice of the pattern)
                clip = binaryTrack.add_clip(0, 0, cboxTr.length, pattern)  #we do nothing with the return value for now.

                #Handle failed to parsed events
                #Add tempo changes to the master calfbox track
                for ev, pulses in cboxTr.failedToParseList:
                    #print (ev[:3])
                    if ev[:3] == [0xFF, 0x51, 0x03]: #tempo signature
                        #print ("Tempo:", ev[:3], ev[3:])
                        #numeric = ev[3] + 256 * ev[4] + 65536 * ev[5]
                        numeric = 65536 * ev[3] + 256 * ev[4] + ev[5]
                        #print (pulses, ":CalfTempo:", float(int(60000000 / numeric)))
                        self.song.set_mti(int(pulses), 8 * float(int(60000000 / numeric))) #TODO: x32 is only because calfbox uses another quarter tick value than laborejo
                    #elif ev[0] & 0xF0 == 0xC0: #patch /instrument change, cut the channel out and see if it is a program change.
                    #    instrument.engine.set_patch(1, 0)

            # Stop the song at the end
            max_pattern_len = max(lengths)
            self.song.set_loop(max_pattern_len, max_pattern_len)
            self.song.update_playback()  #Send the updated song data to the realtime thread
            self.lastMaxTicks = max_pattern_len
            return max_pattern_len

class CalfboxTrack(object):
    """Drop in replacement for libsmf smfTrack"""
    def __init__(self, backendTrack):
        self.pblob = bytes()  # Create a binary blob that contains the MIDI events
        self.length = 0 #set during exportPlayback
        self.failedToParseList = []
        self.backendTrack = backendTrack

    def __deepcopy__(self, memo):
        """Just a precaution. Don't copy anything. Just return itself."""
        return self

    #def add_event(smf.Event(smfEvent), pulses=self.parentWorkspace.score.activeCursor.tickindex + message[1])
    def add_event(self, laborejoEvent, pulses):
        """This is supposed to directly add something on the final
        midi stream.

        Pulses is the start point. It is an int,
        often constructed by the cursor index plus an offset,
        negative or positve

        convertedEvent is in fact not converted but the original
        python data structue. In lismf add_event wants an binary
        encoded event, but we need the raw data."""
        #(384, 0x90, 64, 127),
        #StartTick, type, pitch, velocity
        try:
            if len(laborejoEvent) == 3:
                smfType, param1, param2 = laborejoEvent
            else:
                smfType, param1  = laborejoEvent  #program change has only 2 parameters
                param2 = 0
            #print ("True:", laborejoEvent, pulses)
            self.pblob += cbox.Pattern.serialize_event(pulses, smfType, param1, param2)
        except:
            #print ("False:", laborejoEvent, pulses)
            self.failedToParseList.append((laborejoEvent, pulses))

class States(object):
    def __init__(self):
        """Midi in modes"""
        self.pitchbend = 64 #middle
        self.cc = {
                    0x07:127, #Volume is high
                    0x40:0, #Hold/Sustain Pedal is off.
                    }

    def digitalCC(self, cc):
        if cc in self.cc:
            if self.cc[cc] >= 64:
                return True
            else:
                return False
        else:
            return False
